#include <X11/Xlib.h>
#include <X11/Xproto.h>
#include <X11/extensions/Xfixes.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/shm.h>

#include "services.h"
#include "pvr2d.h"
#include "dri2.h"

typedef Window NativeWindowType;
typedef Display * NativeDisplayType;
typedef Drawable NativePixmapType;

#include "wsegl.h"

typedef struct _wsegldri2_display wsegldri2_display;
struct _wsegldri2_display
{
  unsigned long ref_cnt;
  Display *dpy;
  Bool default_dpy;
  PVR2DCONTEXTHANDLE pvr_context;
  WSEGLConfig *configs;
};

typedef struct _wsegldri2_drawable wsegldri2_drawable;
struct _wsegldri2_drawable
{
  int drawable_type;
  NativePixmapType nativePixmap;
  PVR2DMEMINFO *pvr_meminfo;
  int name;
  void *shmaddr;
  int is_pixmap;
  unsigned int width;
  unsigned int height;
  int pixel_format;
  int stride;
  wsegldri2_display *display;
};


static wsegldri2_display wsegl_display;
static int bpp[] = {2, 2, 4};

static WSEGLError
WSEGLDRI2IsDisplayValid(NativeDisplayType nativeDisplay)
{
  Display *dpy = nativeDisplay;
  WSEGLError rv;

  if (!nativeDisplay)
  {
    nativeDisplay = XOpenDisplay(0);

    if (!nativeDisplay)
      return WSEGL_BAD_NATIVE_DISPLAY;
  }

  if (*DisplayString(nativeDisplay) == ':')
    rv = WSEGL_SUCCESS;
  else
    rv = WSEGL_BAD_NATIVE_DISPLAY;

  if (!dpy)
    XCloseDisplay(nativeDisplay);

  return rv;
}

static WSEGLError
WSEGLDRI2CloseDisplay(WSEGLDisplayHandle handle)
{
  wsegldri2_display *wsegl_dpy = (wsegldri2_display *)handle;

  if (wsegl_dpy->ref_cnt-- == 1)
  {
    PVR2DDestroyDeviceContext(wsegl_dpy->pvr_context);

    if (wsegl_dpy->default_dpy)
      XCloseDisplay(wsegl_dpy->dpy);

    free(wsegl_display.configs);
  }

  return WSEGL_SUCCESS;
}

static WSEGLError
WSEGLDRI2GetDrawableInfo(wsegldri2_display *display, WSEGLConfig *config,
                         WSEGLDrawableHandle *drawable,
                         NativePixmapType nativePixmap,
                         WSEGLRotationAngle *rotationAngle, int drawable_type)
{
  WSEGLError rv;
  wsegldri2_drawable *handle;
  Window window;
  int tmp;
  unsigned int depth;
  unsigned int border_width;
  Status status;

  if (!nativePixmap)
  {
    if (drawable_type == WSEGL_DRAWABLE_WINDOW)
      return WSEGL_BAD_NATIVE_WINDOW;
    else
      return WSEGL_BAD_NATIVE_PIXMAP;
  }

  if (!config || !(drawable_type & config->ui32DrawableType) )
    return WSEGL_BAD_CONFIG;

  handle = (wsegldri2_drawable *)calloc(1, sizeof(*handle));

  if (!handle)
    return WSEGL_OUT_OF_MEMORY;

  handle->drawable_type = drawable_type;
  handle->display = display;
  handle->nativePixmap = nativePixmap;

  if ((status =
       XGetGeometry(display->dpy, nativePixmap, &window, &tmp, &tmp,
                    &handle->width, &handle->height, &border_width, &depth)))
  {
    Bool is_supported;

    if (depth == 24 || depth == 32)
      is_supported = config->ePixelFormat == WSEGL_PIXELFORMAT_8888;
    else
    {
      if (depth != 16)
      {
        rv = WSEGL_BAD_DRAWABLE;
        goto err;
      }

      is_supported = config->ePixelFormat == WSEGL_PIXELFORMAT_565;
    }

    if (is_supported)
    {
      handle->pixel_format = config->ePixelFormat;
      handle->stride = (handle->width + 0x1F) & ~0x1Fu;
      *drawable = handle;
      *rotationAngle = WSEGL_ROTATE_0;

      DRI2CreateDrawable(display->dpy, nativePixmap);

      return WSEGL_SUCCESS;
    }

    rv = WSEGL_BAD_CONFIG;
  }
  else
  {
    rv = WSEGL_BAD_DRAWABLE;
    fprintf(stderr,
            "WSEGLDRI2GetDrawableInfo: Failed to get X drawable geometry - status %d\n",
            status);
    fputs("WSEGL_CreateWindowDrawable: Bad drawable\n", stderr);
  }

err:

  free(handle);

  return rv;
}

static WSEGLError
WSEGLDRI2CreateWindowDrawable(WSEGLDisplayHandle display, WSEGLConfig *config,
                              WSEGLDrawableHandle *drawable,
                              NativeWindowType nativeWindow,
                              WSEGLRotationAngle *rotationAngle)
{
  return WSEGLDRI2GetDrawableInfo((wsegldri2_display *)display, config,
                                  drawable, nativeWindow, rotationAngle,
                                  WSEGL_DRAWABLE_WINDOW);
}

static WSEGLError
WSEGLDRI2CreatePixmapDrawable(WSEGLDisplayHandle display, WSEGLConfig *config,
                              WSEGLDrawableHandle *drawable,
                              NativePixmapType nativePixmap,
                              WSEGLRotationAngle *rotationAngle)
{
  return WSEGLDRI2GetDrawableInfo((wsegldri2_display *)display, config,
                                  drawable, nativePixmap, rotationAngle,
                                  WSEGL_DRAWABLE_PIXMAP);
}

static void
WSEGLDRI2FreeSharedMemory(wsegldri2_drawable *drawable)
{
  PVR2DQueryBlitsComplete(
        drawable->display->pvr_context, drawable->pvr_meminfo, PVR2D_TRUE);
  PVR2DMemFree(drawable->display->pvr_context, drawable->pvr_meminfo);
  shmdt(drawable->shmaddr);
}

static
WSEGLError WSEGLDRI2DeleteDrawable(WSEGLDrawableHandle handle)
{
  wsegldri2_drawable *drawable = (wsegldri2_drawable *)handle;

  DRI2DestroyDrawable(drawable->display->dpy, drawable->nativePixmap);

  if (drawable->pvr_meminfo)
    WSEGLDRI2FreeSharedMemory(drawable);

  free(drawable);

  return WSEGL_SUCCESS;
}

static WSEGLError
WSEGLDRI2SwapDrawable(WSEGLDrawableHandle handle, unsigned long data)
{
  wsegldri2_drawable *drawable = (wsegldri2_drawable *)handle;

  XserverRegion region;
  XRectangle rectangle;

  rectangle.width = drawable->width;
  rectangle.height = drawable->height;
  rectangle.x = 0;
  rectangle.y = 0;
  region = XFixesCreateRegion(drawable->display->dpy, &rectangle, 1);

  DRI2CopyRegion(drawable->display->dpy, drawable->nativePixmap, region, 0, 1);
  XFixesDestroyRegion(drawable->display->dpy, region);
  drawable->is_pixmap = False;

  return WSEGL_SUCCESS;
}

static WSEGLError
WSEGLDRI2SwapControlInterval()
{
  return WSEGL_SUCCESS;
}

static WSEGLError
WSEGLDRI2WaitNative(WSEGLDrawableHandle handle, unsigned long engine)
{
  wsegldri2_drawable *drawable = (wsegldri2_drawable *)handle;

  if (engine != WSEGL_DEFAULT_NATIVE_ENGINE)
    return WSEGL_BAD_NATIVE_ENGINE;

  XSync(drawable->display->dpy, 0);

  return WSEGL_SUCCESS;
}

static WSEGLError
WSEGLDRI2CopyFromDrawable(WSEGLDrawableHandle handle,
                          NativePixmapType nativePixmap)
{
  wsegldri2_drawable *drawable = (wsegldri2_drawable *)handle;
  wsegldri2_display *display = drawable->display;
  int bytes_per_pixel;
  int bits_per_pixel;
  unsigned int red_mask;
  unsigned int green_mask;
  unsigned int blue_mask;
  GC gc;
  XImage image;
  Window window;
  int tmp1;
  unsigned int tmp2;
  unsigned int depth;

  memset(&image, 0, sizeof(image));
  bytes_per_pixel = bpp[drawable->pixel_format];

  if (!XGetGeometry(display->dpy, nativePixmap, &window, &tmp1, &tmp1, &tmp2,
                     &tmp2, &tmp2, &depth))
  {
    return WSEGL_BAD_CONFIG;
  }

  bits_per_pixel = 8 * bytes_per_pixel;

  if (bits_per_pixel != depth )
    return WSEGL_BAD_CONFIG;

  switch (drawable->pixel_format)
  {
    case WSEGL_PIXELFORMAT_4444:
    {
      red_mask = 0xF00;
      green_mask = 0xF0;
      blue_mask = 0xF;
      break;
    }
    case WSEGL_PIXELFORMAT_8888:
    {
      red_mask = 0xFF0000;
      green_mask = 0xFF00;
      blue_mask = 0xFF;
      break;
    }
    default:
    {
      red_mask = 0xF800;
      green_mask = 0x7E0;
      blue_mask = 0x1F;
      break;
    }
  }

  image.red_mask = red_mask;
  image.green_mask = green_mask;
  image.blue_mask = blue_mask;
  image.width = drawable->width;
  image.height = drawable->height;
  image.format = ZPixmap;
  image.bytes_per_line = drawable->stride * bpp[drawable->pixel_format];
  image.data = (char *)drawable->pvr_meminfo;
  image.bitmap_pad = bits_per_pixel;
  image.depth = bits_per_pixel;
  image.bits_per_pixel = bits_per_pixel;
  image.bitmap_unit = bits_per_pixel;

  XInitImage(&image);
  gc = XCreateGC(display->dpy, nativePixmap, 0, 0);
  XPutImage(display->dpy, nativePixmap, gc, &image, 0, 0, 0, 0,
            drawable->width, drawable->height);
  XFreeGC(display->dpy, gc);

  return WSEGL_SUCCESS;
}

static WSEGLError
WSEGLDRI2CopyFromPBuffer(void *address, unsigned long width,
                         unsigned long height, unsigned long stride,
                         WSEGLPixelFormat format, NativePixmapType nativePixmap)
{
  int bytes_per_pixel;
  int bytes_per_line;
  int bits_per_pixel;
  GC gc;
  unsigned long red_mask;
  unsigned long green_mask;
  unsigned long blue_mask;
  XImage image;

  memset(&image, 0, sizeof(image));

  bytes_per_pixel = bpp[format];
  bytes_per_line = stride * bytes_per_pixel;
  bits_per_pixel = 8 * bytes_per_pixel;

  gc = XCreateGC(wsegl_display.dpy, nativePixmap, 0, 0);

  switch (format)
  {
    case WSEGL_PIXELFORMAT_4444:
    {
      red_mask = 0xF00;
      green_mask = 0xF0;
      blue_mask = 0xF;
      break;
    }
    case WSEGL_PIXELFORMAT_8888:
    {
      red_mask = 0xFF0000;
      green_mask = 0xFF00;
      blue_mask = 0xFF;
      break;
    }
    default:
    {
      red_mask = 0xF800;
      green_mask = 0x7E0;
      blue_mask = 0x1F;
      break;
    }
  }

  image.red_mask = red_mask;
  image.green_mask = green_mask;
  image.blue_mask = blue_mask;
  image.depth = bits_per_pixel;
  image.bitmap_pad = bits_per_pixel;
  image.bitmap_unit = bits_per_pixel;
  image.bits_per_pixel = bits_per_pixel;
  image.data = (char *)address + (height - 1) * bytes_per_line;
  image.bytes_per_line = bytes_per_line;
  image.width = width;
  image.height = height;
  image.format = ZPixmap;
  XInitImage(&image);

  image.bytes_per_line = -image.bytes_per_line;

  XPutImage(wsegl_display.dpy, nativePixmap, gc, &image, 0, 0, 0, 0,
            width, height);
  XFreeGC(wsegl_display.dpy, gc);

  return WSEGL_SUCCESS;
}

static WSEGL_FunctionTable const wseglFunctions = {
  WSEGL_VERSION,
  WSEGLDRI2IsDisplayValid,
/*  WSEGLDRI2InitializeDisplay,*/
  .pfnWSEGL_CloseDisplay = WSEGLDRI2CloseDisplay,
  WSEGLDRI2CreateWindowDrawable,
  WSEGLDRI2CreatePixmapDrawable,
  WSEGLDRI2DeleteDrawable,
  WSEGLDRI2SwapDrawable,
  WSEGLDRI2SwapControlInterval,
  WSEGLDRI2WaitNative,
  WSEGLDRI2CopyFromDrawable,
  WSEGLDRI2CopyFromPBuffer,
  /*wseglGetDrawableParameters*/
};

/* Return the table of WSEGL functions to the EGL implementation */
const WSEGL_FunctionTable *
WSEGL_GetFunctionTablePointer(void)
{
    return &wseglFunctions;
}
